--                                                    Rolf Schrder 2000-01-23
--> Includefile GRFUNC.E for function plott. ver. 2.0 Mrkenweg 37
--                                              21029 Hamburg, Germany
--  Language: EUPHORIA ver. 2.2                       r.schr@t-online.de
------------------------------------------------------------------------------
include get.e               -- value()
include graphics.e          -- plott()
include image.e             -- plott()
include machine.e           -- plott()
include misc.e              -- reverse()
include addons\rfont.e      -- select_font(), setxy(), write(),
			    -- set_foreground(), set_background()
------------------------------------------------------------------------------

global function minmax( object x)   -- Minimum and maximum element of x
--              
    atom     mi, ma
    sequence u
--
    if atom(x) then return {x,x} end if
    u = minmax(x[1])
    ma = u[1]
    mi = ma
    for i = 1 to length(x) do
	u = minmax(x[i])
	if ma < u[1] then ma = u[1] end if
	if mi > u[2] then mi = u[2] end if
    end for
    return {mi,ma}
end function
------------------------------------------------------------------------------

function round( atom a, integer n)  -- round atom down to n dec. digits
--       
    sequence s
--
    s = value(sprintf("%" & sprintf("%d.%d",{6+n,n-1}) & "e",a))
    return s[2]
end function
------------------------------------------------------------------------

function make_grid( atom a1, atom a2, integer max) -- rounded number grid
--                                             between a1,a2
    integer  swap, i
    sequence s
    atom     a0, b0, d0
--
    swap = 0
    if    a1 = a2 then
	printf(1,"Grid limits are equal (%g)! Abort by make_grid().\n", a1)
	abort(1)
    elsif a1 > a2 then
	swap = 1
	a0   = a1
	a1   = a2
	a2   = a0
    end if
    d0 = round((a2 - a1)/max,1)
    a0 = floor(a1 / d0) * d0
    i  = 1
    b0 = a1
    s  = {}
    while b0 < a2 do
	b0 = a0 + d0*(i-1)
	s &= b0
	i += 1
    end while
    if swap then
	s  = reverse(s)
    end if
    return s
end function
------------------------------------------------------------------------

global function plott(              -- Screen plott routine
--              
		integer  imode,     -- Graphicsmodus
		sequence gview,     -- Two cornerpoints of pixelwindow
		sequence wkoor,     -- Two cornerpoints of world coordints:
				    --   {{x_left,y_bottom},{x_right,y_top}}
		sequence line,      -- Points of curve/crosses, world coord.:
				    --   {{x1,y1},..{xn,yn}}
		integer  linecolor, -- Linecolor
		integer  textcolor, -- Textcolor
		integer  framecolor,-- Window frame color
		integer  xcolor,    -- Coordinate achses color
		integer  backcolor, -- Window backgroud color
		integer  action)    -- Plottaction:

--  if action = 3 : [set Graphics,] initialize window, plott crosses
--  if action = 2 : [set Graphics,] initialize window, plott curve
--  if action = 1 :                                    plott crosses
--  if action = 0 :                                    plott curve
--  if action =-1 : reset graphics to standard mode, no plott.

    sequence gmode   --  {col?,mod,txtlin,txtclm,xpix,ypix,ncols,pages}
    sequence glin    --  Points of curve/crosses, pixel coord.
    sequence shelp   --  junk
    integer  ix, iy, ixmin, ixmax, iymin, iymax, idx, idy, ihelp
    integer  gviewxmin, gviewxmax, gviewymin, gviewymax
    atom     x0, x1, y0, y1
--
    if action = -1 then
	if graphics_mode(-1) then
	    puts(1,"Error resetting standart graphics. RETURN from plott()\n")
	    return 0
	else
	    return 1
	end if
    end if
    gmode = video_config()
    for i1 = 1 to 2 do      -- order of gview-values, growing from ix1, ix2.
	if gview[1][i1] > gview[2][i1] then
	    ihelp = gview[1][i1]
	    gview[1][i1] = gview[2][i1]
	    gview[2][i1] = ihelp
	end if
    end for
    gview = {gview[1],                  --  complete gview for Polygon
	    {gview[2][1],gview[1][2]},  --
	     gview[2],                  --  gview[1]    gview[2]
	    {gview[1][1],gview[2][2]}}  --
    gviewxmin = gview[1][1]             --  gview[4]    gview[3]
    gviewxmax = gview[2][1]
    gviewymin = gview[1][2]
    gviewymax = gview[4][2]
    x1 = (gview[2][1] - gview[1][1]) / (wkoor[2][1] - wkoor[1][1])
    x0 = 0.5 + gview[1][1] - wkoor[1][1] * x1   -- +0.5 for later use of floor()
    y1 = (gview[4][2] - gview[1][2]) / (wkoor[1][2] - wkoor[2][2])
    y0 = 0.5 + gview[1][2] - wkoor[2][2] * y1   -- +0.5 for later use of floor()
    ixmin = floor( x0 + x1 * line[1][1] )
    ixmax = ixmin
    iymin = floor( y0 + y1 * line[1][2] )
    iymax = iymin
    glin = {}
    for i1 = 1 to length(line) do   -- Transform to pixel coord. for plotting
	ix = floor( x0 + x1 * line[i1][1] )
	iy = floor( y0 + y1 * line[i1][2] )
	glin = append( glin, {ix,iy})
	if glin[i1][1] < ixmin then ixmin = glin[i1][1] end if
	if glin[i1][1] > ixmax then ixmax = glin[i1][1] end if
	if glin[i1][2] < iymin then iymin = glin[i1][2] end if
	if glin[i1][2] > iymax then iymax = glin[i1][2] end if
    end for
    if ixmin < gviewxmin or ixmax > gviewxmax then  -- clip curve to window
	for i1 = 1 to length(line) do               --   in x-direction
	    if glin[i1][1] > gviewxmax then
		glin[i1][1] = gviewxmax
	    elsif glin[i1][1] < gviewxmin then
		glin[i1][1] = gviewxmin
	    end if
	end for
    end if
    if iymin < gviewymin or iymax > gviewymax then  -- clip curve to window
	for i1 = 1 to length(line) do               --   in y-direction
	    if glin[i1][2] > gviewymax then
		glin[i1][2] = gviewymax
	    elsif glin[i1][2] < gviewymin then
		glin[i1][2] = gviewymin
	    end if
	end for
    end if
    if gmode[5] = 0 or gmode[2] != imode then   -- If Textmode or new mode..
	if graphics_mode(imode) then            -- Set Graphicsmodus
	    printf(1,"Can't set GR-mode \"%d\". Return from plott().\n",
		    imode)
	    return 0
	else
	    gmode = video_config()
	end if
    end if
    if action > 1 then                  -- Set new window
	polygon( backcolor, 1, gview )  -- draw backpround
	select_font(ROM8)
	set_foreground(textcolor)
	set_background(CLEAR)           -- = transparent font
	shelp = make_grid(wkoor[1][1], wkoor[2][1], 4)
	for i = 1 to length(shelp) do                       -- X-steps
	    if floor( x0 + x1*shelp[i]) > gviewxmin and
	       floor( x0 + x1*shelp[i]) < gviewxmax then
		if floor(x0+x1*shelp[i]) + 2 +
		   8*(length(sprintf("%g",shelp[i]))) < gviewxmax then
		    setxy(floor(x0+x1*shelp[i]) + 2,gview[4][2] -8)
		    write(sprintf("%g",shelp[i]))
		end if
		draw_line( xcolor,
		{{ floor( x0 + x1*shelp[i]), floor( y0 + y1*wkoor[1][2])},
		{  floor( x0 + x1*shelp[i]), floor( y0 + y1*wkoor[2][2])}})
	    end if
	end for
	shelp = make_grid(wkoor[1][2], wkoor[2][2], 4)
	for i = 1 to length(shelp) do                       -- Y-steps
	    if floor( y0 + y1*shelp[i]) > gviewymin and
	       floor( y0 + y1*shelp[i]) < gviewymax then
		if floor(y0+y1*shelp[i]-9) > gviewymin then
		    setxy(gview[1][1]+2,floor(y0+y1*shelp[i]) -8)
		    write(sprintf("%g",shelp[i]))
		end if
		draw_line( xcolor,
		{{ floor( x0 + x1*wkoor[1][1]), floor( y0 + y1*shelp[i])},
		{  floor( x0 + x1*wkoor[2][1]), floor( y0 + y1*shelp[i])}} )
	    end if
	end for
    end if
    if remainder( action, 2) then           -- Plott crosses or..
	idx = floor(.009*(ixmax-ixmin +.5)) -- half length of crosses in X.
	idy = floor(.012*(iymax-iymin +.5)) -- half length of crosses in Y.
	for i1 = 1 to length(line) do
	    draw_line( linecolor, {{glin[i1][1]-idx, glin[i1][2]    },
				   {glin[i1][1]+idx, glin[i1][2]    }})
	    draw_line( linecolor, {{glin[i1][1],     glin[i1][2]-idy},
				   {glin[i1][1],     glin[i1][2]+idy}})
	end for
    else                                --             ..plott curve.
	draw_line( linecolor, glin)
    end if
    polygon( framecolor, 0, gview )     -- Set frame
    return 0
end function
------------------------------------------------------------------------------
