--                                                    Rolf Schrder 2000-02-22
--> Includefile LSTQ.E fr LSTQ.EX .  ver. 2.00       Mrkenweg 37
--                                              21029 Hamburg, Germany
--  Language: EUPHORIA ver. 2.2                       r.schr@t-online.de
------------------------------------------------------------------------------
include get.e               -- get()
include misc.e              -- PI
include linput.e            -- linput()
------------------------------------------------------------------------------
constant frm1 = "%+13.4e"
constant txt1 = {{"Observations",
		  "Positions",
		  "Weights",
		  "Mean obs.",
		  "Mean errors",
		  "Residuals"},
		 {"Y", "X", "W", "YM", "FYM", "V"}}

sequence X, temp                -- lstq(), runlstq(), mean(), LSTQ-functions
atom     Xi, p0, p1, fij, a1    -- func#()
------------------------------------------------------------------------------

global function read_data(sequence inputfile)
--              
--   Ŀ
--    The sequence: {N,M,Y,X,P} will be returned. Y,X and P are the   
--    sequences of measurements, positions, and weights.              
--    N = equations ( = length of Y, X, P)                            
--    M = unknowns.                                                   
--                                                                    
--    inputfile keeps the data in N+1 lines as lollows:               
--                                                                    
--         1:  M                                                      
--         2:  Y1      X1      P1                                     
--         3:  Y2      X2      P2                                     
--                                                              
--                                                              
--         N:  YN-1    XN-1    PN-1                                   
--       N+1:  YN      XN      PN                                     
--   
    integer  fn, N, M
    sequence xin, yin, pin
    object   line               -- return from linput()

    fn = open( inputfile, "r")
    if fn = -1 then
      puts(1, "Error while opening " & inputfile & ". Abort.\n")
      abort(1)
    end if
    line = linput({},fn)        -- Read first data line
    if integer(line) then
	puts(1,"File " & inputfile & "contains no data! Abort.\n")
	close(fn)
	abort(1)
    end if
    for i = 1 to length(line) do    -- find start of and eliminate comment
	if sequence(line[i]) then
	    if length(line[i]) > 1 then
		if equal("--",line[i][1..2]) then
		    line = line[1..i-1]
		    exit
		end if
	    end if
	end if
    end for
    M = line[1]
    yin = {}
    xin = {}
    pin = {}
    N = 0
    while 1 do
	line = linput({},fn)
	if integer(line) then exit end if
	N += 1
	if length(line) < 3 then
	    printf(1,"Error at line: %d. Abort from linput()\n", N)
	    close(fn)
	    abort(1)
	end if
	yin &= line[1]
	xin &= line[2]
	pin &= line[3]
    end while
    close(fn)
    if N = 0 then
	puts(1,"File " & inputfile & "contains no data! Abort from linput().\n")
	abort(1)
    end if
    return {N, M, yin, xin, pin}  -- read without error.

end function
------------------------------------------------------------------------------

function lstq(              -- LSTQ-routine
--       
	 sequence Y,        -- Observations (measurements)
	 sequence W,        -- Weights of observations
	 integer  F,        -- Func.- identifier of F(i,j) used as matrix F
	 integer  M,        -- Number of unknowns
	 integer  LO)       -- Switch as follows:
--
--  If LO <  0 : return = {W, FWF, AM}
--  If LO =  0 : return = {W, FWF, AM, YM, V, FY}
--  If LO >  0 : return = {W, FWF, AM, YM, V, FY, FAM, FYM}
--
--  The function F(i,j), i=1,N, j=1,M is called via a function identifier.
--  F[i][j]  call_func(F,{i,j}) stands for a NxM-matrix, so that the
--  LSTQ-equations may be written in matrix notation as: Y = FAM + V .
--  The function F(i,j) needs local variables like X (positions of obs.)!
--  The matrix FWF is a M-sequence of M-sequences. (FWF  F'WF), W is
--  a diagonal matrix.
--
    sequence FWF, AM, YM, V, FAM, FYM
    atom     FY             --
    integer  N              -- Number of equations
    integer  in1, in2, in3  -- junk
    atom     at1, at2       -- junk

    N = length(Y)   -- Number of equations
--
--> Normalization of W so that the product of all W[i], (i=1,N) equals 1.
--
    at1 = 1.0
    for i1 = 1 to N do
	at1 *= power( W[i1], 1/N)
    end for
    W /= at1
--
--> Calculate FWF (upper triangle) and FWY (stored in FAM).
--
    FWF = repeat(repeat(0.0,M),M)   -- Initialization MxM-matrix
    YM  = repeat(0.0,M)             -- Initialization (for temp. use)
    FAM = repeat(0.0,M)             -- Initialization (for temp. use)
    AM  = repeat(0.0,M)             -- Initialization (for temp. use)
    for i1 = 1 to N do
	for j1 = 1 to M do
	    YM[j1] = call_func(F,{i1,j1})   -- (temp.)
	end for
	at1 = W[i1]
	FAM += YM * Y[i1] * at1
	for l1 = 1 to M do
	    AM = FWF[l1]
	    at2 = YM[l1] * at1
	    for k1 = l1 to M do
		AM[k1] += YM[k1] * at2
	    end for
	    FWF[l1] = AM
	end for
    end for
--
--> Separate FWF in triangle (Cholesky), store the lower one.
--
    for i1 = 1 to M do
	for j1 = i1 to M do
	    at1 = FWF[i1][j1]
	    if i1 != 1 then
		for k1 = 1 to i1-1 do
		    at1 -= FWF[i1][k1] * FWF[j1][k1]
		end for
	    end if
	    if i1 = j1 then
		if at1 <= 0.0 then
		    puts(1,"Error in lstq.e: Cholesky Deconvolution" &
			   " failed! Abort.\n")
		    abort(1)
		end if
		FWF[i1][i1] = sqrt(at1)
	    else
		FWF[j1][i1] = at1 / FWF[i1][i1]
	    end if
	end for
    end for
--
--> Inverse FWF by use of lower triangle.
--
    for li = 1 to M do
	in1 = 1 + M -li
	for lj = 1 to in1 do
	    in2 = 1 + in1 - lj
	    at2 = 1.0 / FWF[in2][in2]
	    at1 = 0.0
	    if in2 != M then
		if in2 != in1 then
		    in3 = in2 + 1
		    for k1 = in3 to in1 do
			at1 += FWF[k1][in1] * FWF[k1][in2]
		    end for
		end if
		if in1 != M then
		    in3 = in1 + 1
		    for k1 = in3 to M do
			at1 += FWF[in1][k1] * FWF[k1][in2]
		    end for
		end if
		at1 *= -at2
	    end if
	    if in1 = in2 then
		at1 += at2*at2
	    end if
	    FWF[in2][in1] = at1
	end for
    end for
    if M != 1 then
	for i1 = 2 to M do
	    for i2 = 1 to i1 - 1 do
		FWF[i1][i2] = FWF[i2][i1]
	    end for
	end for
    end if
--
--> Calculate AM.
--
    AM = repeat(0.0,M)
    for j1 = 1 to M do
	for k1 = 1 to M do
	    AM[j1] += FWF[j1][k1] * FAM[k1]
	end for
    end for
    if LO < 0 then
	return { W, FWF, AM }
    end if
--
--> Calculate YM, V, FY.
--
    YM  = repeat(0.0,N)                 -- mean of measurements Y
    V   = YM                            -- deviations V = Y - YM
    FYM = YM                            -- Std. deviation of YM
    FY  = 0.0                           -- Std. deviation of single Y(i)
    for i1 = 1 to N do
	for j1 = 1 to M do
	    FAM[j1] = call_func(F,{i1,j1})  -- (temp.)
	    YM[i1] += FAM[j1] * AM[j1]
	end for
	V[i1] = Y[i1] - YM[i1]
	if LO > 0 then
	    at1 = 0.0
	    at2 = FWF[1][1] * FAM[1] * FAM[1]
	    if M != 1 then
		for k1 = 2 to M do
		    for  l1 = 1 to k1 - 1 do
			at1 += FWF[k1][l1] * FAM[k1] * FAM[l1]
		    end for
		    at2 += FWF[k1][k1] * FAM[k1] * FAM[k1]
		end for
	    end if
	    FYM[i1] = sqrt( 2.0 * at1 + at2)
	end if
	FY += W[i1] * V[i1] * V[i1]
    end for
    at1 = N - M
    if at1 = 0.0 then
	at1 = N
    end if
    FY = sqrt( FY / at1)
    if LO = 0 then
	return { W, FWF, AM, YM, V, FY}
    end if
--
--> Calculate FAM, FYM.
--
    for j1 = 1 to M do
	FAM[j1] = FY * sqrt(FWF[j1][j1])
    end for
    FYM *= FY
    return { W, FWF, AM, YM, V, FY, FAM, FYM }
end function
------------------------------------------------------------------------------

global function runlstq(            -- Prepare to run LSTQ
--              
		sequence Y,         -- Observations (measurements)
		sequence xin,       -- Positions of observations
		sequence W,         -- Weights of observations
		integer  F,         -- Functionidentifier (defined in main)
		integer  M,         -- Number of unknowns
		integer  LO)        -- Switch for LSTQ
--
    X = xin             -- Local variable, needed by LSTQ-function for F(i,j)
			-- which will be called from LSTQ routine.
    return { Y, X} & lstq( Y, W, F, M, LO)

end function
------------------------------------------------------------------------------

global function mean(       -- Calculate for given positions xin mean
--                           values YM with std. errors.
		    sequence xin,   -- positions
		    sequence AM,    -- coefficient
		    sequence FWF,   -- covariance matrix of AM
		    integer  F)     -- func. identifier for F(i,j) as matrix F
--
    integer  N, M
    atom     at1, at2
    sequence FAM, YM, FYM

    X    = xin              -- X is modul variable! (for LSTQ-functions)
    N    = length(X)
    M    = length(AM)
    FAM  = repeat(0.0,M)
    YM   = repeat(0.0,N)
    FYM  = YM
    for i1 = 1 to N do
	for j1 = 1 to M do
	    FAM[j1] = call_func(F,{i1,j1})  -- (temp.)
	    YM[i1] += FAM[j1] * AM[j1]
	end for
	at1 = 0.0
	at2 = FWF[1][1] * FAM[1] * FAM[1]
	if M != 1 then
	    for k1 = 2 to M do
		for  l1 = 1 to k1 - 1 do
		    at1 += FWF[k1][l1] * FAM[k1] * FAM[l1]
		end for
		at2 += FWF[k1][k1] * FAM[k1] * FAM[k1]
	    end for
	end if
	FYM[i1] = sqrt( 2.0 * at1 + at2)
    end for
    return { YM, FYM}

end function
------------------------------------------------------------------------------

global procedure write_result( integer fn, sequence result)
--               
    integer N, M
    atom    ajunk, fyfy
--
    N = length(result[1])
    M = length(result[5])
    for i1 = 1 to 2 do
	for i2 = 1 to 6 do
	    printf(fn, "%13s", {txt1[i1][i2]})
	end for
	puts(fn,"\n\n")
    end for
    temp = frm1 & frm1 & frm1 & frm1 & frm1 & frm1 & "\n"
    for i1 = 1 to N do
	printf(fn, temp, {result[1][i1],result[2][i1],result[3][i1],    -- Y, X, P,
			    result[ 6][i1],     -- YM,
			    result[10][i1],     -- FYM,
			    result[ 7][i1]} )   -- V
    end for
    puts  ( 1,"Unknowns AM,\n")
    puts  ( 1,"Std. deviation FAM of AM,\n")
    puts  ( 1,"Std. deviation FY of single measurement Y[i]:\n\n")
    puts  (fn,"\n\n")
    puts  (fn,"Unknowns AM,\n")
    puts  (fn,"Std. deviation FAM of AM,\n")
    puts  (fn,"Std. deviation FY of single measurement Y[i]:\n\n")
    for j1 = 1 to M do
	ajunk = result[5][j1]
	printf( 1, frm1, ajunk )        -- AM
	printf(fn, frm1, ajunk )        -- AM
	if remainder(j1,10)=0 then
	    puts( 1,"\n")
	    puts(fn,"\n")
	end if
    end for
    puts  ( 1,"\n\n")
    puts  (fn,"\n\n")
    for j1 = 1 to M do
	ajunk = result[9][j1]
	printf( 1, frm1, ajunk )        -- FAM
	printf(fn, frm1, ajunk )        -- FAM
	if remainder(j1,10)=0 then
	    puts( 1,"\n")
	    puts(fn,"\n")
	end if
    end for
    puts  ( 1,"\n\n")
    puts  (fn,"\n\n")
    ajunk = result[8]
    printf( 1, frm1, ajunk )            -- FY
    printf(fn, frm1, ajunk )            -- FY
    puts  ( 1,"\n\n")
    puts  (fn,"\n\n")
    puts  ( 1,"Correlations of AM (diagonal = sample variances):\n")
    puts  (fn,"Correlations of AM (diagonal = sample variances):\n")
    fyfy = result[8] * result[8]
    for i1 = 1 to M do
	for j1 = 1 to i1 do
	    if remainder(j1-1,10)=0 then
		puts( 1,"\n")
		puts(fn,"\n")
	    end if
	    if i1 = j1 then
		ajunk = result[4][j1][j1] * fyfy
		printf( 1, frm1, ajunk )
		printf(fn, frm1, ajunk )
	    else
		ajunk = result[4][i1][j1]/
			    sqrt(result[4][i1][i1]*result[4][j1][j1])
		printf( 1, frm1, ajunk )
		printf(fn, frm1, ajunk )
	    end if
	end for
	puts( 1,"\n")
	puts(fn,"\n")
    end for
    puts  ( 1,"\n")
    puts  (fn,"\n")

end procedure
------------------------------------------------------------------------

global function func1(integer I, integer J)
--              
--> function F(I,J) = power(X[I],J)
--
    if J = 1 then
	fij = 1.0
	Xi = X[I]
	return fij
    else
	fij *= Xi/(J-1)
	return fij
    end if

end function
------------------------------------------------------------------------------

global function func2(integer I, integer J)
--              
--> function F(I,J) = L(X[I],J-1) (Leg. Polyn. of order (J-1))
--
    if J = 1 then
	Xi = X[I]
	p1 = 1.0
	fij = 1.0
    else
	p0 = p1
	p1 = fij
	fij = ( (2*J-3)*Xi*p1 - (J-2)*p0 ) / (J-1)
    end if
    return fij

end function
------------------------------------------------------------------------------

global function func3(integer I, integer J)
--              
--> function F(I,J) = T(X[I],J-1)  (Tsch. Polyn. of order (J-1)))
--
    if J = 1 then
	Xi = X[I]
	p1 = Xi
	fij = 1.0
    else
	p0 = p1
	p1 = fij
	fij = 2 * Xi * p1 - p0
    end if
    return fij

end function
------------------------------------------------------------------------------

global function func4(integer I, integer J)
--              
--> function F(I,J) = 1 for J = 1 and for J > 1 :
--                  = cos(X[I]PIfloor(J/2)) for J even,
--                  = sin(X[I]PIfloor(J/2)) for J odd.
--
    if J = 1 then
	return 1.0
    else
	if remainder(J,2) then
	    return sin(Xi)
	else
	    Xi = X[I]*PI*floor(J/2)
	    return cos(Xi)
	end if
    end if

end function

global function func5(integer I, integer J)
--              
--> function F(I,J) = power(X[I],J-1) for J = 1 to 4 and
--                  = cos(X[I]PI]    for J = 5,
--                  = sin(X[I]PI]    for J = 6.
--
    if J = 1 then
	Xi = X[I]
	fij = 1.0
	return 1.0
    elsif J < 5 then
	fij *= Xi
	return fij
    elsif J = 5 then
	return cos(Xi*PI)
    else
	return sin(Xi*PI)
    end if

end function
------------------------------------------------------------------------------

global function func6(integer I, integer J)
--              
--> function F(I,J) = power(X[I],J-1) for J = 1 to 4 and
--                  = cos(X[I]PIfloor((J-3)/2)) for J > 4 and odd,
--                  = sin(X[I]PIfloor((J-3)/2)) for J > 4 and even.
--
    if J = 1 then
	fij = 1.0
	Xi = X[I]
	return fij
    elsif J < 5 then
	fij *= Xi
	return fij
    elsif remainder(J,2) then  -- J = 5 then
	a1 = Xi*PI*floor((J-3)/2)
	return cos(a1)
    else
	return sin(a1)
    end if

end function
------------------------------------------------------------------------------
