--                                                    Rolf Schrder 2001-02-28
--> LSTQ.EX   Linear Least Square Solution  Ver. 1.2  Mrkenweg 37
--                                             21029 Hamburg, Germany
--  Language: EUPHORIA ver. 2.2                       r.schr@t-online.de
------------------------------------------------------------------------------
-- without type_check
include graphics.e                  -- text_rows()
include image.e                     -- get_all_palette(), save_screen()
include linput.e                    -- linput()
include grfunc.e                    -- minmax(), plott(),  and so on ...
include lstq.e                      -- lstq(),
                                    -- read_data(),
                                    -- runlstq(),
                                    -- write_result(),
                                    -- mean(),
                                    -- and the following LSTQ-functions:
------------------------------------------------------------------------------
constant func = { "A1+A2X+A3X^2/2!+...+AMX^(M-1)/(M-1)!",            -- 1
                  "A1L(X,0)+A2L(X,1)+A3L(X,2)+...+AML(X,M-1) " &    -- 2
                    "(Leg. Polyn.)",
                  "A1T(X,0)+A2T(X,1)+A3T(X,2)+...+AMT(X,M-1) " &    -- 3
                    "(Tsch. Polyn.)",
                  "A1+[j=2,M] Sj; with arg=floor(j/2)X it " &       -- 4
                    "follows for Sj:\n      if j even then " &
                    "Sj=AjCOS(arg) else Sj=AjSIN(arg).",
                  "A1+A2X+A3X+A4X^3+A5COS(X)+A6SIN(X) ",      -- 5
                  "[j=1,4] AjX^(j-1) + [j=5,M] Sj; with " &          -- 6
                    "arg=floor((j-3)/2)X it \n      " &
                    "follows for Sj: if j odd then Sj=AjCOS(arg) " &
                    "else Sj=AjSIN(arg)."}
------------------------------------------------------------------------------
constant DOSGRMODE = { 261, 260, 18, 19}    -- Standard graphics modes
constant DOSWINDOW = {
                      {{   0,   0}, {1023, 767}},   -- 261
                      {{   0,   0}, {1023, 767}},   -- 260
                      {{   0,   0}, { 639, 479}},   --  18
                      {{   0,   0}, { 319, 199}}    --  19
                     }
------------------------------------------------------------------------------

function test_grmode(sequence modi)
--       
    integer m

    for i = 1 to length(modi) do
        if graphics_mode(modi[i]) = 0 then
            m = graphics_mode(-1)
            return i
        end if
    end for
    puts(1,"Can't set any GR-mode. Abort\n")
    return 0
end function
------------------------------------------------------------------------------

function plottlstq(         -- plott-function called from main().
--                 -- Returns 0 if no plot possible else 1.
                    integer  mode,      -- graphics mode
                    sequence gview,     -- graphics window (pixel)
                    integer  InitFlag,  -- if InitFlag = TRUE then initialize
                    sequence result,    -- else add plott           [ plott
                    integer  F,
                    integer  func_num)
--
    sequence wcoor          --  Corner points of world coord. in pixelwindow
    sequence xp             --  x-dots for curve plott
    sequence line           --  Points of curve or crosses
    integer  linecolor      --  Color of curve
    integer  textcolor      --  Textcolor
    integer  framecolor     --  Framecolor of window
    integer  xcolor         --  Color of coordinate axes
    integer  backcolor      --  Backcolor of window
    atom     wscale         --  Reduce wkoor-window by wscale in pixel-window
    atom     xm, ym, dx, dy --  Mean and half width of X- and Y- range
    atom     a1, a2, fyfy   --  junk
    sequence Xlim, Ylim     --  junk
    integer  np, ip         --  junk
    sequence X, Y
    integer  N
--
    Y = result[1]
    X = result[2]
    N = length(Y)
--
    framecolor = 7      -- 7, 8
    xcolor     = 0      -- 0
    backcolor  = 8      -- 8, 15
    textcolor  = 2      -- 2
    wscale     = 0.95   -- 0.95   part of pixel-window used as wcoordinates
    Xlim = minmax(X)
    Ylim = minmax(Y)
    if func_num = 4 then     -- Fourier
        Xlim[2] += (Xlim[2]-Xlim[1])/(length(X)-1)
    end if
    xm = 0.5*(Xlim[2]+Xlim[1])
    dx = 0.5*(Xlim[2]-Xlim[1])
    ym = 0.5*(Ylim[2]+Ylim[1])
    dy = 0.5*(Ylim[2]-Ylim[1])
    if dy = 0 then      -- No plott possible!
        return 0
    end if
    wcoor = {{xm-dx/wscale,ym-dy/wscale},{xm+dx/wscale,ym+dy/wscale}}
--
--  Compute np X-positions xp, the mean yn(xpi) and the std. dev. fyni.
--
    np = gview[2][1] - gview[1][1]   -- (plott-dots -1) in x-direction
    np = floor(np*wscale)
    a2  = (Xlim[2]-Xlim[1])/(np-1)
    a1  = Xlim[1]-a2
    xp   = repeat(0.0,np)
    for i = 1 to np do
      xp[i] = a1 + a2*i
    end for
    fyfy = result[8] * result[8]
    result[4] *= fyfy       -- Estimated Covariancematrix.
    result = mean( xp, result[5], result[4], F)
    line = repeat({0.0,0.0},N)
    for i = 1 to N do
      line[i][1] =  X[i]
      line[i][2] =  Y[i]
    end for
    linecolor  = 1      -- Cross color: blue=1, bright blue=9
    if InitFlag then
        ip = plott( mode,   -- initialize cross graph, plott observations
                    gview,
                    wcoor,
                    line,
                    linecolor,
                    textcolor,
                    framecolor,
                    xcolor,
                    backcolor,
                    3)
    else
        ip = plott( mode,   -- plott observations
                    gview,
                    wcoor,
                    line,
                    linecolor,
                    textcolor,
                    framecolor,
                    xcolor,
                    backcolor,
                    1)
    end if
    line = repeat({0.0,0.0},np)
    for i = 1 to np do
      line[i][1] = xp[i]
      line[i][2] = result[1][i] + result[2][i]
    end for
    linecolor = 3       -- Std. +error color: cyan=3, brown=6
    ip = plott( mode,   -- curve graph
                gview,
                wcoor,
                line,
                linecolor,
                textcolor,
                framecolor,
                xcolor,
                backcolor,
                0)
    for i = 1 to np do
      line[i][2] = result[1][i] - result[2][i]
    end for
    linecolor = 3       -- Std. -error color: cyan=3, brown=6
    ip = plott( mode,
                gview,
                wcoor,
                line,
                linecolor,
                textcolor,
                framecolor,
                xcolor,
                backcolor,
                0)
    for i = 1 to np do
      line[i][2] = result[1][i]
    end for
    linecolor = 12      -- Mean value color: 12=red
    ip = plott( mode,
                gview,
                wcoor,
                line,
                linecolor,
                textcolor,
                framecolor,
                xcolor,
                backcolor,
                0)
    return 1
end function
------------------------------------------------------------------------------

procedure main()
--        
    sequence command, infile, outfile, bmpfile, datin, gmode, sjunk
    sequence X, Y, W, result
    integer  N, M, LO, F, ip, fn, func_num
    integer  imode
    atom     t0
--
    gmode = video_config()
    N = text_rows(50)
    puts(1,"Linear Least Square Solution (LSTQ) for the following functions:\n")
    puts(1, repeat('',79)&"\n")
    printf(1," %2d = %-19s\n", { 0, "Abort program"})
    for i1 = 1 to length(func) do
        printf(1," %2d = %-19s\n", { i1, func[i1]})
    end for
    puts(1, repeat('',79)&"\n")
    puts(1," Selection           : ")
    func_num = linput({},0)
    puts(1,"\n")
    if func_num < 1 then
        N = text_rows(gmode[3])
        abort(1)
    end if
    sjunk = "func" & sprintf("%d",func_num)
    puts(1, repeat('',79)&"\n\n")
    infile  = "LSTQ.DAT"        -- Standart Inputfile (Data)
    outfile = "LSTQ.OUT"        -- Standart Outputfile
    bmpfile = "LSTQ.BMP"        -- Standart Bitmapfile
    command = command_line()
    if length(command) > 2 then -- are filenames given with command?
        infile = command[3]
        if length(command) > 3 then
            outfile = command[4]
            if length(command) > 4 then
                bmpfile = command[5]
            end if
        end if
    end if
    fn = open( outfile, "w")
    t0 = time()
    datin = read_data( infile )             -- read data
    t0 = time() - t0
    N = datin[1]
    M = datin[2]
    Y = datin[3]
    X = datin[4]
    W = datin[5]
    F = routine_id(sjunk)                   -- function for designmatrix.
    if F = -1 then
        puts(1,"Can't find selected function \"" & sjunk & "\" =\n"  &
                                                 func[func_num] & ". Abort.\n")
        puts(1,"press any key...")
        ip = wait_key()
        N = text_rows(gmode[3])
        abort(1)
    end if
    printf( 1,"Equations N = %d\nUnknowns  M = %d\n\n", {N,M})
    printf(fn,"Equations N = %d\nUnknowns  M = %d\n\n", {N,M})
    printf( 1,"Data reading: %6.2fsec\n\n", t0)
    printf(fn,"Data reading: %6.2fsec\n\n", t0)
    t0 = time()
--
--> result, depending on LO:
--
--  result : { Y, X, W, FWF, AM, YM, V, FY, FAM, FYM }
--
-- LO < 0  :  
-- LO = 0  :  
-- LO > 0  :  
--
    LO = 1  -- calculate ALL in lstq()
    result = runlstq( Y, X, W, F, M, LO)
    t0 = time()-t0
    write_result(fn, result)
    printf( 1,"Calculation: %6.2fsec\n", t0)
    printf(fn,"Calculation: %6.2fsec\n", t0)
    close( fn )
    puts(1,"press any key...")
    ip = wait_key()
    N = text_rows(gmode[3])
    use_vesa(1)
    imode = test_grmode(DOSGRMODE)
    imode = plottlstq(DOSGRMODE[imode],
                      DOSWINDOW[imode], 1, result, F, func_num)
    if imode then
--      Save the screen as BMP-file
        ip = save_screen( 0, bmpfile)
        ip = wait_key()
        ip = graphics_mode(-1)
    else
        ip = graphics_mode(-1)
        printf(1,"No plot created. All input data values are equal %+13.4e\n",
                                                        {result[5][1]})
        puts(1,"press any key...")
        ip = wait_key()
    end if
end procedure
------------------------------------------------------------------------------

main()
------------------------------------------------------------------------------

