--                                                    Rolf Schrder 2000-02-27
--> Includefile LSTQD.E for LSTQD.EX .                Mrkenweg 37
--                                             21029 Hamburg, Germany
--  Language: EUPHORIA ver. 2.2                       r.schr@t-online.de
------------------------------------------------------------------------------
include sort.e              -- sort()
include machine.e           -- makefile(): set_rand()
include misc.e              -- PI
include poisson.e           -- poisson()
------------------------------------------------------------------------------
constant MAXRND = #3FFFFFFF     -- func7() and gauss(): rand()
constant FAKRND = 1/#40000000   -- func7() and gauss(): rand()
--
sequence X                  -- makefile() and LSTQ-functions
atom     Xi, p0, p1, fij,   -- func#(), makefile()
	 a1, a2             --
------------------------------------------------------------------------------

global function func1(integer I, integer J)
--              
--> function F(I,J) = power(X[I],J)
--
    if J = 1 then
	Xi = X[I]
	fij = 1.0
    else
	fij *= Xi/(J-1)
    end if
    return fij

end function
------------------------------------------------------------------------------

global function func2(integer I, integer J)
--              
--> function F(I,J) = L(X[I],J-1) (Leg. Polyn. of order (J-1))
--
    if J = 1 then
	Xi = X[I]
	p1 = 1.0
	fij = 1.0
    else
	p0 = p1
	p1 = fij
	fij = ( (2*J-3)*Xi*p1 - (J-2)*p0 ) / (J-1)
    end if
    return fij

end function
------------------------------------------------------------------------------

global function func3(integer I, integer J)
--              
--> function F(I,J) = T(X[I],J-1)  (Tsch. Polyn. of order (J-1)))
--
    if J = 1 then
	Xi = X[I]
	p1 = Xi
	fij = 1.0
    else
	p0 = p1
	p1 = fij
	fij = 2 * Xi * p1 - p0
    end if
    return fij

end function
------------------------------------------------------------------------------

global function func4(integer I, integer J)
--              
--> function F(I,J) = 1                       for J = 1, else:
--                  = cos(X[I]PIfloor(J/2)) for J > 1 and even,
--                  = sin(X[I]PIfloor(J/2)) for J > 1 and odd.
--
    if J = 1 then
	return 1.0
    else
	if remainder(J,2) then
	    return sin(Xi)
	else
	    Xi = X[I]*PI*floor(J/2)
	    return cos(Xi)
	end if
    end if

end function
------------------------------------------------------------------------------

global function func5(integer I, integer J)
--              
--> function F(I,J) = power(X[I],J-1) for J = 1 to 4 and
--                  = cos(X[I]PI]    for J = 5,
--                  = sin(X[I]PI]    for J = 6.
--
    if J = 1 then
	Xi = X[I]
	fij = 1.0
	return fij
    elsif J < 5 then
	fij *= Xi
	return fij
    elsif J = 5 then
	return cos(Xi*PI)
    else
	return sin(Xi*PI)
    end if

end function
------------------------------------------------------------------------------

global function func6(integer I, integer J)
--              
--> function F(I,J) = power(X[I],J-1) for J = 1 to 4 and
--                  = cos(X[I]PIfloor((J-3)/2)) for J > 4 and odd,
--                  = sin(X[I]PIfloor((J-3)/2)) for J > 4 and even.
--
    if J = 1 then
	Xi = X[I]
	fij = 1.0
	return 1.0
    elsif J < 5 then
	fij *= Xi
	return fij
    elsif remainder(J,2) then       -- J = 5,7,9,... then
	a1 = Xi*PI*floor((J-3)/2)
	return cos(a1)
    else                            -- J = 6,8,10,...
	return sin(a1)
    end if

end function
------------------------------------------------------------------------------

global function func7(integer I, atom sigma)
--              
--> Random values from Poisson-density with expectation l.
--
    return poisson(sigma)
end function
------------------------------------------------------------------------------
global function func8(integer I, integer J)
--              
--> Random values from Gau-density with expectation zero und std. dev. 1.
--
    return sqrt(-2.0*log(FAKRND*rand(MAXRND))) * cos(PI*FAKRND*rand(MAXRND))
end function
------------------------------------------------------------------------------

global function func9(integer I, integer J)
--              
--> return: 1 for |X[I]| < , -1 else.
--
    if X[I] < -0.5 or X[I] > 0.5 then
	return -1.0
    else
	return 1.0
    end if
end function
------------------------------------------------------------------------------

global function func10(integer I, integer J)
--              
--> return: SIG( X[I] )
--
    return (X[I] > 0.0) - (X[I] < 0.0)
end function
------------------------------------------------------------------------------

function gauss(atom sigma)
--       
--> Random values from Gau-density with expectation zero und std. dev. sigma.
--
    return sigma * sqrt(-2.0*log(FAKRND*rand(MAXRND))) *
		    cos(PI*FAKRND*rand(MAXRND))
end function
------------------------------------------------------------------------------

global procedure makefile(  -- Writes data file for LSTQ-program.
--               
		    sequence filenam,
		    integer  func_num,
		    integer  func_id,
		    sequence A,
		    atom     Xmin,
		    atom     Xmax,
		    integer  N,
		    integer  M,
		    atom     sigma,
		    integer  LO)
--
    integer  fn
    atom     t0, sum
    sequence now
--
    fn = open( filenam, "w")
    printf(fn, "%10d    %s%d\n", {M,"-- = Unknowns. Function No: ", func_num})
    now = date()
    t0 = 3600*now[4]+60*now[5]+now[6]   -- time in seconds.
    set_rand(t0)                        -- Initialisaton of rand()
    X = repeat(MAXRND,N)
    if LO = 1 then                  -- Random Xi-distribution!
	if func_num = 4 then
	    X = Xmin + rand(X)*FAKRND*(Xmax-Xmin)*(N-1)/N
	else                        -- for all other functions
	    X = Xmin + rand(X)*FAKRND*(Xmax-Xmin)
	end if
	X = sort(X)
    else                            -- Aequidistant Xi-distribution!
	if func_num = 4 then        -- Fourier-Reihe
	    a2 = (Xmax-Xmin)/N      -- Xi = a1 + a2*i ; X1 = Xmin,
	    a1 = Xmin - a2          --                  XN = Xmin+Xmax(N-1)/N
	else                        -- for all other functions
	    a2 = (Xmax-Xmin)/(N-1)  -- Xi = a1 + a2*i ; X1 = Xmin, XN = Xmax
	    a1 = Xmin - a2
	end if
	for i1 = 1 to N do
	    X[i1] = a1 + a2 * i1
	end for
    end if
    if    func_num = 9 or           -- Rectangle
	  func_num = 10 then         -- Signum
	for i1 = 1 to N do
	    sum = call_func(func_id,{i1,1}) + gauss(sigma)
	    printf(fn, "%+25.15e%+25.15e%+25.15e\n", { sum, X[i1], 1.0 })
	end for
    elsif func_num = 7 then         -- Poisson
	for i1 = 1 to N do
	    sum = call_func(func_id,{i1,sigma})
	    printf(fn, "%+25.15e%+25.15e%+25.15e\n", { sum, X[i1], 1.0 })
	end for
    elsif func_num = 8 then         -- Gau
	for i1 = 1 to N do
	    sum = call_func(func_id,{i1,1}) * sigma
	    printf(fn, "%+25.15e%+25.15e%+25.15e\n", { sum, X[i1], 1.0 })
	end for
    else
	for i1 = 1 to N do
	    sum = 0.0
	    for j1 = 1 to M do
		sum += A[j1]*call_func(func_id,{i1,j1})
	    end for
	    sum += gauss(sigma)
	    printf(fn, "%+25.15e%+25.15e%+25.15e\n", { sum, X[i1], 1.0 })
	end for
    end if
    close( fn )
end procedure
------------------------------------------------------------------------------
