--                                                    Rolf Schrder 1999-12-26
--> Includefile POISSON.E               ver 2.00      Mrkenweg 37
--                                           21029 Hamburg, Germany
--  Language: EUPHORIA ver. 2.2                       r.schr@t-online.de
------------------------------------------------------------------------------
include machine.e                   -- set_rand()
include misc.e                      -- PI
------------------------------------------------------------------------------
constant MAXRND = power(2,30) - 1           -- rand():
constant FAKRND = power(2,-30)              --   (0,1)  FAKRND*rand(MAXRND)
constant cE     = 2.718281828459045         -- exp(x) = power(cE,x)
constant SQ2PI  = 2.5066282746310005        -- gamma(), lngamma()
constant C0     = 1.000000000190015         -- gamma(), lngamma()
constant cof    = {  7.618009172947146e+1,  -- gamma(), lngamma()
                    -8.650532032941677e+1,
                     2.401409824083091e+1,
                    -1.231739572450155e+0,
                     1.208650973866179e-3,
                    -5.395239384953e-6}
------------------------------------------------------------------------------
    atom oldm, g, sq, logl                  -- poisson()
    sequence now                            -- for date() to initialize rand()

    oldm = -1.0                             -- flag for poisson()
    now  = date()                           -- get time for set_rand()
    set_rand(3600*now[4]+60*now[5]+now[6])  -- Initialisaton of rand()
------------------------------------------------------------------------------\

global function exp(object x)   -- Exponential function
--              
    return power(cE, x)
end function
------------------------------------------------------------------------------

global function gamma(object x)    -- Gamma function ( || < 2/10^10 )
--              
    atom     y, tmp, sum
    sequence s

    if atom(x) then
        if x < 0.0 then
            return PI/(gamma(1.0-x)*sin(PI-PI*x))
        end if
        tmp = x + 5.5
        tmp = power(tmp, x + 0.5) * power(cE, -tmp) * SQ2PI
        y   = x
        sum = C0
        for i = 1 to 6 do
            y   += 1.0
            sum += cof[i]/y
        end for
        return tmp * sum / x
    else
        s = repeat(0.0,length(x))
        for i = 1 to length(x) do
            s[i] = gamma(x[i])
        end for
        return s
    end if
end function
------------------------------------------------------------------------------

global function lngamma(object x)    -- Logarithmus of gamma function
--                             ( || < 2/10^10 )
    atom     y, tmp, sum
    sequence s

    if atom(x) then
        if x < 1.0 then
            return lngamma(x+1.0) - log(x)
        end if
        y   = x
        tmp = x + 5.5
        tmp = (x + 0.5) * log(tmp) - tmp
        sum = C0
        for i = 1 to 6 do
            y   += 1.0
            sum += cof[i]/y
        end for
        return tmp + log(SQ2PI*sum/x)
    else
        s = repeat(0.0,length(x))
        for i = 1 to length(x) do
            s[i] = lngamma(x[i])
        end for
        return s
    end if
end function
------------------------------------------------------------------------------

global function poisson(object l) -- Random values from Poisson distribution.
--                       f(k) := l^k / (k!exp(l)). l = Expectation(k).
--
    atom     p, t, y
    sequence s

    if atom(l) then
        if l < 12.0 then
            if l != oldm then
                oldm = l
                g    = power(cE,-l)
            end if
            p  = 0.0
            t  = FAKRND*rand(MAXRND)
            while t > g do
                p += 1.0
                t *= FAKRND*rand(MAXRND)
            end while
        else
            if l != oldm then
                oldm = l
                sq   = sqrt(2.0*l)
                logl = log(l)
                g    = l * logl - lngamma(l+1.0)
            end if
            t = 0.0
            while FAKRND*rand(MAXRND) > t do
                y = tan(PI*FAKRND*rand(MAXRND))
                p = sq * y + l
                if p < 0.0 then
                    t = 0.0
                else
                    p = floor(p)
                    t = 0.9*(1.0+y*y)*power(cE, p*logl - lngamma(p+1.0) - g)
                end if
            end while
        end if
        return p
    else
        s = repeat(0,length(l))
        for i = 1 to length(l) do
            s[i] = poisson(l[i])
        end for
        return s
    end if
end function
------------------------------------------------------------------------------
